import functools
from .Terminal import *

class List:
    """
    Type safe collection with custom functions
    Includes:
    - reduce, filter, map
    """
    def __init__(self, type) -> None:
        self.contents = list()
        self.type = type
        pass

    def append(self, value):
        if isinstance(value, self.type):
            self.contents.append(value)
        else:
            Terminal.error(f"Type mismatch:{type(value)} is not {self.type}")
            raise TypeError

    def extend(self, iterator):
        for i in iterator:
            self.append(i)
    
    def filter(self, fn):
        self.contents = list(filter(fn, self.contents))

    def map(self, fn):
        self.contents = list(map(fn, self.contents))

    def reduce(self, fn) -> any:
        return functools.reduce(fn, self.contents)
    
    def count(self) -> int:
        return len(self.contents)

    def first(self):
        try: 
            return self.contents[0]
        except:
            return None
    
    def forEach(self, fn): 
        for i in self.contents: fn(i)
    
    def __str__(self) -> str:
        text = "" + "["
        spacer = ", "
        for i in self.contents:
            text += i + spacer
        text = text.removesuffix(spacer)
        text += "]"
        return text


class Date:
    def __init__(self, dateString, origin) -> None:
        self._dateString = dateString
        self._origin = origin
        self.date = None
    
    def __str__(self) -> str:
        return self.date


URL_PREFIX = "https://www."


class href:
    def __init__(self, string, origin) -> None:
        self.content = string
        self.origin = origin
    
    def url(self): 
        url = URL_PREFIX + self.origin + self.content
        try:
            scheme = self.content.split(":")[0]
            if scheme == "https": return self.content
            else: return url
        except:
            return url

    def __str__(self) -> str:
        return self.url()


from .Response import *
from .RequestError import *
from .Core.List import *
from .ResponseType import *

class Responses(List):
    def __init__(self) -> None:
        List.__init__(self, ResponseType)

    def removeErrors(self):
        """
        Use to filter a list to only have responses without errors
        """
        self.filter(self._removeErrors)
    
    def _removeErrors(self, response):
        return isinstance(response, Response)


from .Fetcher import Fetcher
from .Scraper import Scraper
from .RequestError import RequestError
from .Core.Terminal import Terminal

class Crawler:
    def __init__(self, root) -> None:
        self.root = root
        self.fetcher = Fetcher()


    def crawl(self):
        response = self.fetcher.get_page(self.root)
        if self.isError(response): 
            Terminal.error("Error at root")
            return
        responses = self._crawling_(response)
        self._crawl(responses)

    
    def _crawl(self, responses):
        for response in responses.contents:
            urls = self._crawling_(response)
            if urls: self._crawl(urls)
    
    def _crawling_(self, response):
        scraper = Scraper(response)
        list_href = scraper.gethref()
        if list_href.count == 0 or not list_href: return None
        list_href.map(lambda x: x.url())
        list_href.forEach(lambda x: Terminal.found(x))
        responses = self.fetcher.get_all_pages(list_href)
        responses.removeErrors()
        return responses
    
    def isError(self, response): return isinstance(response, RequestError)




from .Core.Terminal import *
from .ResponseType import *

class Response(ResponseType):
    def __init__(self, text, url) -> None:
        self.text = text
        self.url = url
        self.origin = self.getOrigin()
        self.debug_init()
    
    def getOrigin(self):
        try: 
            split = self.url.split(".")
            domain = split[1]
            TLD = split[2].split("/")[0]
            return domain + "." + TLD
        except:
            return None
    
    def debug_init(self):
        Terminal.put(f"{self.url} - {bcolors.OKGREEN}success{bcolors.ENDC}")
    
    def __str__(self) -> None:
        return f"Response({self.url[0:10]})"


import requests
from .RequestError import *
from .Response import *
from .Responses import *



class Fetcher:
    """
    Fetch a url and its contents
    methods:
    - get_page(url: String) -> Response
    - get_all_pages(urls: [String]) -> Responses
    """
    def __init__(self) -> None:
        pass

    def get_page(self, url):
        """
        Retrive a single page
        """
        r = requests.get(url)
        status_code = r.status_code
        html_content = r.text
        return Response(html_content, url) if status_code == 200 else RequestError(status_code, url)

    def get_all_pages(self, urls):
        """
        Retrive all pages
        """
        contents = Responses()
        for url in urls:
            response = self.get_page(url)
            contents.append(response)
        return contents


from bs4 import BeautifulSoup
from .Topic import *
from .Core.List import *
from .Core.Terminal import *
from .href import href

BS4_FEATURES = "lxml"

class Scraper:
    """
    Param:
        response: Response
    Attributes:
        title: str
        topic: [Topic]
        date: datetime
        url: str
    """
    def __init__(self, resonpse) -> None:
        self.response = resonpse
        self.topic = List(Topic)
        self.date = None
        self.url = resonpse.url
        self.origin = self.response.origin
        self._soup = BeautifulSoup(resonpse.text, features = BS4_FEATURES)
        self.title = self.getTitle()

    def getTitle(self) -> str:
        """
        Get The title of a webpage
        """
        for title in self._soup.find_all("title"):
            return title.text
            
    
    def getTopics(self):
        """
        Get topics of a webpage
        """
        pass
    
    def getDate(self):
        """
        Gets date of webpage
        """
        pass

    
    def clean(self):
        """
        Remove all style and js tags
        """
        # parse html content
    
        for data in self._soup(['style', 'script']):
            # Remove tags
            data.decompose()
    
        # return data by retrieving the tag content
        return ' '.join(self._soup.stripped_strings)

    def gethref(self) -> List:
        """
        Returns a list of links
        """
        hrefList = List(href)
        hrefList.extend([href(a['href'], self.origin) for a in self._soup.find_all('a', href = True)])
        return hrefList

    def __str__(self):
        return f"""
        title = {self.title}
        topic = {self.topic}
        date = {self.date}
        url = {self.url}
        origin = {self.origin}
        """




from .Core.Terminal import *
from .ResponseType import *

class RequestError(ResponseType):
    def __init__(self, status_code, url) -> None:
        self.status_code = status_code
        self.url = url
        self.debug_init()
    
    def debug_init(self):
        Terminal.put(f"{self.url} - {bcolors.FAIL}ERROR{bcolors.ENDC}")
    
    def __str__(self) -> str:
        return f"RequestError({self.url[0:10]}, {self.status_code})"



import pathlib
from setuptools import setup, find_packages
from version import Version

v = Version()

HERE = pathlib.Path(__file__).parent

VERSION = '0.1.' + str(v.version)
PACKAGE_NAME = 'sloom'
AUTHOR = 'Reed Hunsaker'
AUTHOR_EMAIL = 'reed.hunsaker@gmail.com'
URL = 'https://github.com/ReedHunsaker/Webscraper'

LICENSE = 'Apache License 2.0'
DESCRIPTION = 'Webscraper'
LONG_DESCRIPTION = (HERE / "README.md").read_text()
LONG_DESC_TYPE = "text/markdown"

INSTALL_REQUIRES = [
      'requests',
      'bs4',
      'pytest'
]

setup(name=PACKAGE_NAME,
      version=VERSION,
      description=DESCRIPTION,
      long_description=LONG_DESCRIPTION,
      long_description_content_type=LONG_DESC_TYPE,
      author=AUTHOR,
      license=LICENSE,
      author_email=AUTHOR_EMAIL,
      url=URL,
      install_requires=INSTALL_REQUIRES,
      packages=find_packages()
      )




class bcolors:
    HEADER = '\033[95m'
    OKBLUE = '\033[94m'
    OKCYAN = '\033[96m'
    OKGREEN = '\033[92m'
    WARNING = '\033[93m'
    FAIL = '\033[91m'
    ENDC = '\033[0m'
    BOLD = '\033[1m'
    UNDERLINE = '\033[4m'

class Terminal:
    def __init__(self) -> None:
        pass

    @staticmethod
    def put(s):
        print(f"> {s}")
    
    @staticmethod
    def error(s):
        print(f"> {bcolors.FAIL}ERROR{bcolors.ENDC} - {s}")
    
    @staticmethod
    def found(s):
        print(f"> {bcolors.OKGREEN}URL FOUND:{bcolors.ENDC} {s}")